from bohra.launcher.Utils import CustomFormatter, _check_path, _get_required_columns, _check_size_file, _check_input_snippy, _check_reference, _check_mask, _compartive_args
import pandas as pd
import pathlib
import os
import logging


# Logger
LOGGER =logging.getLogger(__name__) 
LOGGER.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(CustomFormatter())
fh = logging.FileHandler('bohra.log')
fh.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%Y-%m-%d %I:%M:%S %p') 
fh.setFormatter(formatter)
LOGGER.addHandler(ch) 
LOGGER.addHandler(fh)

def _get_annotation(input_file:str, mtb:bool) -> str:
    annot_cols =[]
    if _check_path(input_file):
        df = pd.read_csv(input_file, sep='\t')
        cols = df.columns.tolist()
        std_cols = _get_required_columns()
        std = []
        for col in std_cols:
            for c in std_cols[col]:
                std.append(c)
        
        for c in cols:
            if c not in std:
                val = df[c].unique().tolist()
                if len(val) == 1 and val[0] == "not_supplied":
                    continue
                annot_cols.append(c)
    if mtb:
        annot_cols.extend(['Phylogenetic lineage', "Predicted drug resistance"])
    return ','.join(annot_cols) if annot_cols else ""

def _setup_comparative_args(kwargs:dict, command:dict, mtb:bool) -> dict:
    if mtb:
        command['modules'].append('mtb')
    command['modules'].append(kwargs['comparative_tool'])
    command['params'].append(f"--tree_input {kwargs['tree_input'] if kwargs['comparative_tool'] != 'mash' else 'distance'}")
    command['modules'].append(f"{kwargs['tree_builder'] if kwargs['comparative_tool'] != 'mash' else 'quicktree'}")
    command['params'].append(f"--cluster {'false' if kwargs['comparative_tool'] == 'mash' else kwargs['cluster']}")
    command['params'].append(f"--ignore_warnings {'true' if kwargs['ignore_warnings'] else 'false'}")
    
    if kwargs['comparative_tool'] == "snippy" and not _check_input_snippy(kwargs['input_file']):
        LOGGER.critical("You have selected snippy as your tool for snp detection. Currently this requires paired-end reads. Please check your input file and try again.")
        raise SystemExit
    # command = _accessory_params(kwargs=kwargs, command=command)
    annots = _get_annotation(kwargs['input_file'], mtb)
    ref = _check_reference(kwargs['reference_genome'], kwargs['workdir'])
    mask = _check_mask(kwargs['mask'])
    if annots != "":
        command['params'].append(f"--annot_cols '{annots}'")
    command['params'].append(f"--reference {ref}")
    command['params'].append(f"--mask {mask}")
    if ref == "no_ref" and kwargs['comparative_tool'] == "snippy":
        LOGGER.critical("You have selected snippy as your tool for snp detection. Currently this requires a reference genome. Please check your input file and try again.")
        raise SystemExit
    command = _compartive_args(tool = kwargs['comparative_tool'],kwargs=kwargs, command=command)
        
    if 'gubbins' in kwargs and kwargs["gubbins"]:
        command['params'].append("--gubbins true")
    
    return command


#!/usr/bin/env python3

import sys,pandas as pd

tool = sys.argv[1]
data = sys.argv[2]
df = pd.read_csv(data, sep="\t")

if tool == "sylph":
    sp = df["Species"].values[0]
else:
    sp = df["Match 1"].values[0]


print(sp)


#!/usr/bin/env python3

import pandas as pd
import sys



def combine_stats(dist_df, stats_df) -> pd.DataFrame:
    dist_df = dist_df.rename(columns = {"Match count":"# shared kmers"})
    # print(dist_df)
    tmp1 = stats_df.rename(columns = {'Isolate': 'Sample1', '# kmers': 'Sample1 #kmers'})
    # print(tmp1)
    tmp2 = stats_df.rename(columns = {'Isolate': 'Sample2', '# kmers': 'Sample2 #kmers'})
    # print(tmp2)
    merged = dist_df.merge(tmp1[['Sample1', 'Sample1 #kmers']], on = 'Sample1', how = 'left')
    # print(merged)
    merged = merged.merge(tmp2[['Sample2', 'Sample2 #kmers']], on = 'Sample2', how = 'left')
    # print(merged)
    merged["Sample1 % contribution"] = round(merged['# shared kmers'] / merged['Sample1 #kmers'] * 100, 2)
    merged["Sample2 % contribution"] = round(merged['# shared kmers'] / merged['Sample2 #kmers'] * 100, 2)
    merged["Distance"] = merged["Distance"].apply(lambda x: int(x))
    # print(merged)
    merged["Minimum % contribution"] = merged[["Sample1 % contribution", "Sample2 % contribution"]].min(axis=1)
    merged = merged[[ "Sample1", "Sample1 #kmers", "Sample2", "Sample2 #kmers", "# shared kmers", "Sample1 % contribution", "Sample2 % contribution", "Minimum % contribution","Distance"]]
    print(merged)
    merged.to_csv("ska_distance_stats.tsv", sep = '\t', index = False)
    return merged
def open_stats(ska_nk):

    try:

        stats = pd.read_csv(ska_nk, sep = '\t')
        return stats
    except Exception as e:
        print(f"Error reading SKA nk stats file: {e}")
        sys.exit(1)
def make_matrix(df, col1, col2, vals) -> pd.DataFrame:
    """
    Create a matrix from two columns of a DataFrame and a list of values.

    Args:
        df (pd.DataFrame): The input DataFrame.
        col1 (str): The name of the first column.
        col2 (str): The name of the second column.
        vals (list): A list of values to include in the matrix.

    Returns:
        pd.DataFrame: A DataFrame representing the matrix.
    """
    idx = sorted(set(df[col1]).union(df[col2]))
    # Create a pivot table
    matrix = df.pivot(index = col1, columns = col2, values = vals).reindex(index = idx, columns = idx).fillna(0, downcast = 'infer').pipe(lambda x: x+x.values.T)
    # Reindex to include all values
    matrix = matrix.reset_index()
    matrix = matrix.rename(columns = {col1: 'Isolate'})

    return matrix


dist_stats = sys.argv[1]
df = pd.read_csv(dist_stats, sep = '\t')
ska_nk = sys.argv[2]

stats = open_stats(ska_nk)
distance_metrics = combine_stats(df, stats)
matrix = make_matrix(df, 'Sample1', 'Sample2', 'Distance')
matrix.to_csv("distances.tsv", sep = '\t', index = False)


#!/usr/bin/env python3
import sys, json, datetime, subprocess, pathlib,pandas


def get_vals_seqtk(f):

    with open(f, 'r') as inputfile:
        lines = inputfile.read().split('\n')
        qscore = lines[1].split('\t')[3]
        bases = lines[1].split('\t')[0]
        gc= float(lines[1].split('\t')[1]) + float(lines[1].split('\t')[2])
        try:
            return round(float(qscore),1)
        except ValueError:
            print(f"Something has gone horribly wrong {qscore} can not be parsed to a float!!")
            return 0.0
        
def sum_stats(f,cmd,col):

    p = subprocess.run(f"cat {f} | datamash {cmd} --header-in {col}", shell = True, capture_output = True, encoding= 'utf-8')

    return p.stdout

def get_vals_kit(f):

    reads = sum_stats(f, 'sum', 4)
    minval = sum_stats(f, 'mean', 6)
    avgval = sum_stats(f, 'mean', 7)
    maxval = sum_stats(f, 'mean', 8)

    return int(reads.strip()), float(minval.strip()), float(avgval.strip()), float(maxval.strip())

def get_length(genome_size):

    with open(genome_size, 'r') as f:
        x = f.read().strip().split('\n')[0].strip()
        print(x)
        try:
            return int(x)
        except:
            print(f"Something has gone horribly wrong {x} can not be parsed to an int!!")
            return 0
    
def get_dpth(genome_size,bases):
    length = get_length(genome_size)
    try:
        dpth = int(bases)/length
        return round(dpth, 1),int(length)
    except:
        print(f"Something has gone horribly wrong with the genome size {genome_size} {bases} can not be parsed to an int!!")
        return 0.0, int(length)
    
def get_gcs(path):

    try:
        gcs = pandas.read_csv(path, sep = '\t')
        return gcs[gcs.columns[0]].values[0]
    except Exception as e:
        print(f"Something has gone wrong with the GC content file {path} {e}")
        return 0.0

tab = pandas.read_csv(sys.argv[2], sep = '\t')
tab['Isolate'] = sys.argv[1]
print(tab)
gcs = get_gcs(sys.argv[3])
print(gcs)
print(sys.argv[6])
tab = tab.rename(columns = {'num_seqs': 'Reads', 'sum_len': 'Yield','min_len':'Min len','max_len':'Max len', 'avg_len':'Avg len', 'Q30(%)': 'Qscore'})
dpth,size = get_dpth(genome_size = sys.argv[4], bases = tab['Yield'].values[0])
tab['Depth'] = dpth
tab["Genome size"] = size
tab['GC'] = gcs
tab['is_control'] = True if "control" in sys.argv[7] else False
tab['filesize'] = "<20000000" if sys.argv[8] == "FAIL_READ_FILE_TOO_SMALL" else ">20000000"
tab['Qscore'] = get_vals_seqtk(sys.argv[5])

tab = tab[['Isolate','Reads','Yield','GC','Min len','Avg len','Max len','Qscore',"Genome size", 'Depth',"is_control","filesize"]]
tab.to_csv('read_assessment.txt', sep = '\t', index = False)





    





from bohra.launcher.Utils import CustomFormatter, _check_path, _run_subprocess, _get_required_columns, _get_columns_list
from bohra.launcher.BohraSetupFiles import _open_input_file, _check_data_format, _make_workdir
from bohra.launcher.BohraSetupResources import _get_config, _max_cpus
from bohra.launcher.BohraBasic import _setup_basic_args
from bohra.launcher.BohraAssembly import _setup_assembly_args
from bohra.launcher.BohraTyping import _setup_typing_args
from bohra.launcher.BohraComparative import _setup_comparative_args
from bohra.launcher.BohraPangenome import _setup_pangenome_args
from bohra.launcher.Deps import dependencies

import pandas as pd
import sys
import pathlib
import os
import logging
import datetime
import subprocess

# Logger
LOGGER =logging.getLogger(__name__) 
LOGGER.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(CustomFormatter())
fh = logging.FileHandler('bohra.log')
fh.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%Y-%m-%d %I:%M:%S %p') 
fh.setFormatter(formatter)
LOGGER.addHandler(ch) 
LOGGER.addHandler(fh)



def _setup_working_directory(input_file:str,
                             workdir : str) -> bool:

    input_data = _check_data_format(df = _open_input_file(input_file))

    _make_workdir(_input=input_data, workdir=workdir)
    
    
    return True

def _init_command_dict(profile:str, cpus:int, job_name:str, prefix:str, pipeline:str, profile_config:str, trim:bool, report_outdir:str, outdir: str, no_downloadable_tables: bool) -> dict:

    command_dict = {"params":[
        f"--report_outdir {report_outdir}",
        f"-profile {profile}",
        f"--outdir {outdir}",
        # f"--no_downloadable_tables {'true' if no_downloadable_tables else 'false'}",
        f"--pipeline {pipeline}",
        f"-executor.cpus {cpus}",
        "-with-trace",
        f"--job_id {job_name}",
        # f"--conda_prefix {prefix}",
                ], "modules":[]}

    if profile_config != "":
        command_dict["params"].append(f"-c {profile_config}")
    if trim:
        command_dict["modules"].append(f"trim")

    return command_dict

def _funcs() -> dict:
    """Returns a dictionary of functions to be used in the pipeline."""
    
    

    return {
        "basic":[],
        "assemble": [_setup_assembly_args],
        "amr_typing":[ _setup_typing_args],
        "full":[_setup_typing_args, _setup_comparative_args, _setup_pangenome_args ],
        "comparative":[ _setup_comparative_args ],
        "preview":[ _setup_comparative_args ],
        "tb":[ _setup_comparative_args ],
    }

def _make_command(command: dict) -> str:
    """Constructs the command string from the command dictionary."""
    
    cmd = f"nextflow -Dnxf.pool.type=sync run {pathlib.Path(__file__).parent.parent.resolve() / 'bohra.nf'} " 
    
    cmd += " ".join(command["params"])
    if command["modules"]:
        cmd += " --modules " + ",".join(command["modules"])
    
    return cmd

def _generate_bohra_command(
        pipeline: str,
        kwargs: dict) -> str:
    # print(kwargs)
    if _make_workdir(workdir=kwargs["workdir"],   
                  _input=kwargs["input_file"], 
                  report_outdir=kwargs["report_outdir"], 
                  replace_report=kwargs["replace_report"]):
        LOGGER.info(f"Checking on the setup for the {pipeline} pipeline.")
        if not kwargs["no_check"]:
            dependencies(_action = "check")
            
        max_cpus = int(_max_cpus(cpus=kwargs.get('cpus', 0)))
        LOGGER.info(f"Using {int(max_cpus)} CPUs for the {pipeline} pipeline.")
        profile,profile_config = _get_config(user_config=kwargs['profile_config'], title = kwargs['job_name'],cpus=max_cpus, wd = kwargs["workdir"])
        command = _init_command_dict(profile=profile, profile_config = profile_config, cpus=max_cpus, job_name = kwargs.get('job_name', 'bohra'), prefix=kwargs.get('conda_prefix', 'bohra'), pipeline=pipeline, trim = kwargs["trim"], report_outdir=kwargs["report_outdir"], outdir=kwargs['workdir'], no_downloadable_tables=kwargs['no_downloadable_tables'])
        
        # update kwargs with checked input file
        kwargs["input_file"] = "input_checked.tsv"
        command["params"].append(f"--isolates {kwargs['input_file']}")
        LOGGER.info(f"Input file {kwargs['input_file']} added to command successfully.")
        command = _setup_basic_args(kwargs=kwargs, command=command, pipeline=pipeline)
        mtb = True if pipeline == "tb" else False
        LOGGER.info(f"Setting up arguments for the {pipeline} pipeline.")
        for _func in _funcs()[pipeline]:
            command = _func(kwargs=kwargs, command=command, mtb = mtb)
        # command = _funcs()[pipeline](kwargs=kwargs, command=command)
        
        cmd = _make_command(command=command)
        return cmd
    else:
        LOGGER.error(f"Failed to create the working directory {kwargs['workdir']}.")
        raise SystemError
        
def _run_bohra_cmd(
        cmd: str) -> subprocess.CompletedProcess:
    
    LOGGER.info(f"Running the command: {cmd}")
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    for line in proc.stdout:
        print(line, end='')

    # Wait for the process to complete and get the return code
    proc.wait()
    return proc
def _check_bohra_success(expected_output: str) -> bool:
    LOGGER.info(f"Checking for the expected output: {expected_output}")
    if pathlib.Path(expected_output).exists():
        LOGGER.info(f"The bohra run has completed successfully.")
        return True
    else:
        LOGGER.critical(f"The bohra run failed to generate the expected output: {expected_output}.")
        return False

def run_bohra(
        pipeline: str,
        kwargs: dict) -> bool:
    
        # generate the command
        cmd = _generate_bohra_command(pipeline=pipeline, kwargs=kwargs)
        # if no_auto_run is set, just print the command OR run it
        if kwargs["no_auto_run"]:
            LOGGER.info(f"Please paste the following command to run the pipeline:\n\033[1m{cmd}\033[0m")
        else:
            proc = _run_bohra_cmd(cmd=cmd)
            # check for success
            if not _check_bohra_success(expected_output=pathlib.Path(kwargs['report_outdir']) / f"{kwargs['job_name']}.html"):
                LOGGER.critical(f"Bohra pipeline failed.")
                raise SystemExit(1)
            else:
                return True


#!/usr/bin/env python3

import pandas as pd
import sys



pth = sys.argv[1]
out_stats = sys.argv[2]

with open(pth, 'r') as f:
    lines = f.read().strip().split('\n')

union = int([line for line in lines if line.startswith('k-mers=')][0].split('=')[1].strip())
names = eval([line for line in lines if line.startswith('sample_names=')][0].split('=')[1])
numkmers = eval([line for line in lines if line.startswith('sample_kmers=')][0].split('=')[1])

raw = dict(zip(names, numkmers))

data = []

for r in raw:
    d = {
        'Isolate': r,
        '# kmers': raw[r],
        'Union # kmers': union,
        'Proportion in union (%)': round((raw[r] / union) * 100, 2)
    }

    data.append(d)

df = pd.DataFrame(data)
df.to_csv(out_stats, sep = '\t', index = False)


#!/usr/bin/env python3
import sys,  json
import pandas as pd


species = sys.argv[1]

_files = sys.argv[2:]

for f in _files:
    df = pd.read_csv(f, sep='\t')
    df["Species"] = species
    # out_file = f.replace('.txt', f'_{species}_matches.txt')
    df.to_csv(f, sep='\t', index=False)


#!/usr/bin/env python3
import pathlib, subprocess, sys, pandas, numpy,json

import pandas, pathlib


_file = sys.argv[1]

with open(_file, 'r') as j:
    _d = json.load(j)
    header = ['Isolate','Scheme','ST']
    row = [_d[0]['id'],_d[0]['scheme'],_d[0]['sequence_type']]
    try:
        als = sorted(list(_d[0]['alleles'].keys()))
        for a in range(len(als)):
            header.append(f'Allele {a + 1}')
            row.append(f"{als[a]}({_d[0]['alleles'][als[a]]})")
    
        h = '\t'.join(header)
        r = '\t'.join(row)
    
        pathlib.Path('mlst.txt').write_text('\n'.join([h,r]))
    except:
        pathlib.Path('mlst.txt').write_text('\n'.join(['\t'.join(header), '\t'.join(row)]))


"""Bohra 
.. moduleauthor:: Kristy Horan <kristyhoran15@gmail.com>

Bohra exinct tree kangaroo that lived on the nullarbor

Bohra is microbial genomics pipeline, designed predominantly for use in public health, but may also be useful in research settings. The pipeline takes as input a tab-delimited file with the isolate IDs followed by the path to READ1 and READ2, a reference for alignment and a unique identifier, where reads are illumina paired end reads (other platforms are not supported).

The pipline is based on nullarbor (https://github.com/tseemann/nullarbor) and is designed to be run in high performance computing environment.

Bohra is modular allowing the user to choose between calling SNPs and generating a phylogenetic tree, perform assemblies and detect AMR, perform typing etc; Or use the full pipeline to call SNPs, generate phylogenies, assemble, type and detect the pan-geneome. The output of Bohra is a html report that can be distributed, with downloadable tables and data.
"""


# import logging
# import argparse
import click
import sys

from bohra.version import version
from click.exceptions import UsageError
from click._compat import get_text_stderr

from bohra.launcher.Utils import _get_run_cmd_options, _get_dep_cmd_options, _get_util_options
from bohra.commands.init_databases import init_databases
# from bohra.commands.install import install 
from bohra.commands.generate_input import generate_input
from bohra.commands.test import bohratest
# from bohra.commands.check_setup import check_setup
from bohra.launcher.BohraRun import run_bohra
from bohra.launcher.Deps import dependencies


def _show_usage_error(self, file=None):
    if file is None:
        file = get_text_stderr()
    color = None
    if self.ctx is not None:
        color = self.ctx.color
        click.echo(self.ctx.get_help() + '\n', file=file, color=color)
    click.echo('ERROR: %s' % self.format_message(), file=file, color=color)

UsageError.show = _show_usage_error
CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help'],
    
)
@click.group(no_args_is_help=True, context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.version_option(version, message='%(prog)s %(version)s')
# @click.pass_context
def cli():
    pass

@cli.group(no_args_is_help=True, context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
def run():
    """
    Run the Bohra pipeline.
    """
    pass


@cli.group(no_args_is_help=True,invoke_without_command=True)
def deps():
    """
    Manage bohra dependencies.
    """
    pass

@cli.group(no_args_is_help=True, context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
def utils():
    """
    Bohra utilities.
    """
    pass


def generic_options(_func: click.Command, options: dict) -> click.Command:
    """Add generic options to a Click command."""
    for opt in options:
        if 'short_name' in opt:
            click.option(
                f"--{opt['name']}",
                f"{opt['short_name']}",
                type=opt.get('type', None),
                default=opt.get('default', None),
                help=opt.get('help', ''),
                show_default= True,
                is_flag= opt.get('is_flag', False),
            )(_func) # Apply the decorator to the function
        else:
            click.option(
                f"--{opt['name']}",
                type=opt.get('type', None),
                default=opt.get('default', None),
                help=opt.get('help', ''),
                show_default= True,
                is_flag= opt.get('is_flag', False),
            )(_func)
    return _func
    

def create_utils_subcommand_with_options(name, options_dict):
    f"""Dynamically created a subcommand with options from a list."""

    @utils.command(name=name, help = f"Help for {name} utility. COMING SOON.")
    def utils_subcommand(**kwargs):
        # try:
        #     if dependencies(_action=name, kwargs=kwargs) == 0:
        #         click.echo(f"Successfully completed {name}ing utilities.")
        #     else:
        #         raise UsageError(f"An error.")
        # except Exception as e:
        #     raise SystemExit
        pass

    utils_subcommand = generic_options(utils_subcommand, options_dict)
    return utils_subcommand

def create_deps_subcommand_with_options(name, options_dict):
    f"""Dynamically created a subcommand with options from a list."""

    @deps.command(name=name, help = f"Help for {name}ing dependencies.")
    def deps_subcommand(**kwargs):
        try:
            if dependencies(_action=name, kwargs=kwargs) == 0:
                click.echo(f"Successfully completed {name}ing dependencies.")
            else:
                raise SystemExit(1)
        except Exception as e:
            raise SystemExit(1)


    deps_subcommand = generic_options(deps_subcommand, options_dict)
    return deps_subcommand


def create_run_subcommand_with_options(name, options_dict):
    f"""Dynamically created a subcommand with options from a list."""

    @run.command(name=name, help = f"Help for the {name} pipeline.")
    def run_subcommand(**kwargs):
        # try:
        if len(sys.argv) > 3:
            run_bohra(pipeline=name, kwargs=kwargs)
        else:
            raise UsageError(f"You did not supply any input. Please supply some input to run the {name} pipeline.")
        # except Exception as e:
        #     raise UsageError(f"An error occurred while running the {name} pipeline: {e}")

    # for opt in options_dict:
        
    #     if 'short_name' in opt:
    #         click.option(
    #             f"--{opt['name']}",
    #             f"{opt['short_name']}",
    #             type=opt.get('type', None),
    #             default=opt.get('default', None),
    #             help=opt.get('help', ''),
    #             show_default= True,
    #             is_flag= opt.get('is_flag', False),
    #         )(run_subcommand) # Apply the decorator to the function
    #     else:
    #         click.option(
    #             f"--{opt['name']}",
    #             type=opt.get('type', None),
    #             default=opt.get('default', None),
    #             help=opt.get('help', ''),
    #             show_default= True,
    #             is_flag= opt.get('is_flag', False),
    #         )(run_subcommand)
    run_subcommand = generic_options(run_subcommand, options_dict)
    return run_subcommand

run_cmd_opts = _get_run_cmd_options()
dep_cmd_opt = _get_dep_cmd_options()
util_cmd_opt = _get_util_options()
for opt in run_cmd_opts:
    create_run_subcommand_with_options(name = opt, options_dict = run_cmd_opts[opt])

for opt in dep_cmd_opt:
    create_deps_subcommand_with_options(name = opt, options_dict = dep_cmd_opt[opt])

for opt in util_cmd_opt:
    create_utils_subcommand_with_options(name = opt, options_dict = util_cmd_opt[opt])

cli.add_command(init_databases.init_databases)
# cli.add_command(install.install_deps)
# cli.add_command(check_setup.check_deps)
cli.add_command(generate_input.generate_input)
cli.add_command(bohratest.test)



if __name__ == '__main__':
    cli()

